from ..config import config
from .objs import Record, Array, Atomic
from .inference.reduce import reduce_schema


def fit(data):
    if isinstance(data, dict):
        schema_content = dict()
        for key in data:
            schema_content[key] = fit(
                data[key]
            )
        schema = Record(schema_content)
        if config.unify_records:
            schema = try_unify_dict(schema)
    elif isinstance(data, list):
        schema = Array(reduce_schema([fit(e) for e in data]))
    elif data is None:
        schema = Atomic(None)
    else:
        schema = Atomic(type(data))
    return schema


def try_unify_dict(dict_schema):
    uni_dict = dict_schema.to_uniform_dict()
    if isinstance(uni_dict._content, Record) or isinstance(
            uni_dict._content, Array):
        return uni_dict
    else:
        return dict_schema



"""
Record Json Schema Objects

TODO:
- [X] Allow another DynamicDict to capture mandatory fields & co-occurrence relationship between fields
- [X] Enable representing the __init__ of DynamicDict (same as other schema types)
"""
from __future__ import annotations
import copy
from collections import Counter
from ...config import config
from ..inference.reduce import reduce_schema
from .basic import JsonSchema

__all__ = [
    'Record',
    'UniformRecord',
    'DynamicRecord'
]


class Record(JsonSchema):
    def __init__(self, content: dict):
        super().__init__(content)

    def check_content(self):
        assert isinstance(self._content, dict), 'Record content should be dict'
        for key in self._content:
            assert isinstance(key, str), 'Record content key should be str'
            assert isinstance(
                self._content[key], JsonSchema), 'Record content value should be JsonSchema'

    def __repr__(self):
        return f'Record({self._content})'

    def __hash__(self):
        return hash(tuple(sorted(self._content.items())))

    def __or__(self, e):
        if isinstance(e, DynamicRecord):
            return e | self
        elif isinstance(e, Record):
            old = copy.deepcopy(self)
            new = copy.deepcopy(e)
            if old._content.keys() == new._content.keys():
                old = Record.merge_label_equal_fields(old, new)
                return old
            else:
                if config.equivalence_mode == 'kind':
                    return DynamicRecord.merge_records_as_dynamic_record(
                        old, new)
                elif config.equivalence_mode == 'label':
                    return self._base_or(e)
        else:
            return self._base_or(e)

    @staticmethod
    def merge_label_equal_fields(old: Record, new: Record):
        for key in old._content:
            old._content[key] |= new._content[key]
        return old

    def to_uniform_dict(self):
        schemas = [v for v in self._content.values()]
        uniform_content = reduce_schema(schemas)
        return UniformRecord(uniform_content)


class DynamicRecord(Record):
    """
    Dictionary where keys are not strict
    (some keys can be optional)
    """

    def __init__(self, content: dict, key_counter):
        super().__init__(content)
        self._key_counter = key_counter

    def __repr__(self):
        return f'DynamicRecord({self._content}, {self._key_counter})'

    def __hash__(self):
        return hash(tuple(sorted(self._content.items()))) + \
            hash(tuple(sorted(self._key_counter.items())))

    def __or__(self, e):
        if isinstance(e, DynamicRecord):
            old = copy.deepcopy(self)
            new = copy.deepcopy(e)
            return DynamicRecord.merge_dynamic_records(old, new)
        elif isinstance(e, Record):
            old = copy.deepcopy(self)
            new = copy.deepcopy(e)
            return DynamicRecord.merge_dynamic_n_normal_records(old, new)
        else:
            return self._base_or(e)

    @staticmethod
    def merge_dynamic_n_normal_records(old: DynamicRecord, new: Record):
        result_dict = DynamicRecord.__merge_common_fields(old, new)
        for key in new._content.keys():
            old._key_counter[key] += 1
        return DynamicRecord(result_dict, old._key_counter)

    @staticmethod
    def merge_dynamic_records(old: DynamicRecord, new: DynamicRecord):
        result_dict = DynamicRecord.__merge_common_fields(old, new)
        return DynamicRecord(
            result_dict, old._key_counter + new._key_counter)

    @staticmethod
    def merge_records_as_dynamic_record(old: Record, new: Record):
        key_counter: Counter = Counter()
        result_dict = DynamicRecord.__merge_common_fields(old, new)
        for key in old._content.keys():
            key_counter[key] += 1
        for key in new._content.keys():
            key_counter[key] += 1
        return DynamicRecord(result_dict, key_counter)

    @staticmethod
    def __merge_common_fields(old: Record, new: Record):
        result_dict = {}
        old_keys = list(old._content.keys())
        new_keys = list(new._content.keys())
        all_keys = set(old_keys + new_keys)
        for key in all_keys:
            if key in old._content and key in new._content:
                result_dict[key] = old._content[key] | new._content[key]
            elif key in old._content:
                result_dict[key] = old._content[key]
            elif key in new._content:
                result_dict[key] = new._content[key]
        return result_dict


class UniformRecord(JsonSchema):
    """
    Dictionary where value elements
    are united into a JsonSchema.
    """

    def __init__(self, content: JsonSchema):
        super().__init__(content)

    def check_content(self):
        assert isinstance(
            self._content, JsonSchema), f'UniformRecord content should be JsonSchema, but it is {self._content}'

    def __repr__(self):
        return f'UniformRecord({self._content})'

    def __or__(self, e):
        if isinstance(e, UniformRecord):
            new = copy.deepcopy(e)
            old = copy.deepcopy(self)
            return UniformRecord(old._content | new._content)
        else:
            return self._base_or(e)



import execnet
import inspect
from functools import wraps
exec('from ..schema.objs import *')
exec('from collections import Counter')


def build_remote(gw, function, engine='pypy'):
    func_name = function.__name__
    consumer_str = f"""
{inspect.getsource(function)}
args, kwargs = channel.receive()
channel.send(repr({func_name}(*args, **kwargs)))
"""
    consumer_str = consumer_str.replace(f'@{engine}', '')
    channel = gw.remote_exec(consumer_str)
    return channel


def wrap(func, channel):
    @wraps(func)
    def wrapped_func(*args, **kwargs):
        channel.send((args, kwargs))
        return eval(channel.receive())
    return wrapped_func


def pypy(func):
    gw = execnet.makegateway("popen// pypy3")
    channel = build_remote(gw, func, engine='pypy')
    wrapped_func = wrap(func, channel)
    return gw, wrapped_func


def python(func):
    gw = execnet.makegateway("popen// python")
    channel = build_remote(gw, func, engine='python')
    wrapped_func = wrap(func, channel)
    return gw, wrapped_func



"""
Config the Json Schema Inference Model
"""


class Config:
    """
    - 1. equivalence model: kind | label
        - kind: Turn on DynamicRecord for unifing Records with different fields
        - label: Turn off DynamicRecord and construct Union of Records with inconsistent fields
    - 2. enable_uniform_record: True | False (whether try to unify Record values for Record with values of same kind)
    """

    def __init__(self, unify_records=True, equivalence_mode='kind'):
        self.init(
            unify_records=unify_records,
            equivalence_mode=equivalence_mode)

    def init(self, unify_records=True, equivalence_mode='kind'):
        self._unify_records = unify_records
        assert equivalence_mode == 'kind' or equivalence_mode == 'label'
        self._equivalence_mode = equivalence_mode

    @property
    def unify_records(self) -> bool:
        return self._unify_records

    @property
    def equivalence_mode(self) -> str:
        return self._equivalence_mode


config = Config()
init = config.init



import argparse
import autopep8
from jsonschema_inference.inference import JsonlInferenceEngine


def run() -> None:
    parser = argparse.ArgumentParser(
        description='Inferencing Json Schema')

    parser.add_argument('--jsonl',
                        type=str, required=True,
                        help="Inference Json Schema from .jsonl file")

    parser.add_argument('--nworkers',
                        type=int, required=False, default=1,
                        help="Inference Worker Count")

    parser.add_argument('--verbose',
                        type=bool, required=False, default=False,
                        help="Showing the Result by Pretty Print")

    parser.add_argument('--out',
                        type=str, required=False, default='',
                        help="Saving the json schema into a output file")

    args = parser.parse_args()
    print(f"Your json file is at: {args.jsonl}")
    print('verbose:', args.verbose)

    class Engine(JsonlInferenceEngine):
        @property
        def jsonl_path(self):
            return args.jsonl

    schema = Engine(
        inference_worker_cnt=args.nworkers).get_schema(
        verbose=args.verbose)
    schema_str = autopep8.fix_code(str(schema))
    if args.verbose:
        print(schema_str)
    if args.out != '':
        with open(args.out, 'w') as f:
            f.write(schema_str)



"""
A Json schema builder from multiple strutured Json files

NOTE:

What is a structured JSON file?

It is a dictionary with key and elements

Elements can be `List`, `Dictionary`, or `Simple Variables`, which is a value or string (Optional[int], Optional[float], ...).

If the element is a `List`, it must have elements of same `Simple Variables` or `Dictionary` with same keys.

If the element is a `Dictionary`, it should have fixed keys and fixed type of element.

If the element is a `Simple Variables`, it should have fixed type.

TODO:
- [X] Build another create_schema with keys of dictionary shown
- [X] Generate consistet schema to allow multiple json's schema to be merged into a more consistent schema
- [X] Convert Schema to a Python DataClass
    - [X] A class extract json properties one-by-one
    - [X] A class generate overall json content
- [X] An adaptor that takes json as input and initialize the python DataClass

- [-] Change redis-server cache to another one


TODO:
- [X] Use Python Cuckoo Filter to avoid repeat download of json
- [ ] enable no dump mode

REF:
https://github.com/huydhn/cuckoo-filter
https://dl.acm.org/doi/pdf/10.1145/2674005.2674994

"""
import abc
import os
import pickle
import math
import logging
import signal
import sys
import typing
import requests
import tqdm
from multiprocessing.pool import ThreadPool
import json as json_package
from ..schema.objs import JsonSchema
from ..schema import InferenceEngine

__all__ = ['APIInferenceEngine']


class APIInferenceEngine:
    """
    Args:
        - api_thread_cnt: number of threads downloading json from url
        - inference_worker_cnt: number of processes inferencing the json schema
        - json_per_worker: number of json files an inference worker takes as input
        - cuckoo_dump: path to a dump file to store the record of processed index
        - schema_dump: path to a dump to store the inferenced json schema.
    Methods to be overide:
        - index_generator: a generator yeilding index (or url) strings referencing to a json file
        - index_to_url: a function takes the index from index_generator as input and convert it to an url
            (If index_generator yeilds url, no need to overide this function.)
        - is_json_valid: a function takes determine whether a json is valid or not
            (The invlid json would be ignored.)
    """

    def __init__(self, api_thread_cnt=1000, inference_worker_cnt=4, json_per_worker=1000,
                 cuckoo_dump='cuckoo.pickle', schema_dump='schema.pickle', jsonl_dump=None):
        self._api_thread_cnt = api_thread_cnt
        self._inference_worker_cnt = inference_worker_cnt
        if self._inference_worker_cnt > 1:
            from ray.util.multiprocessing import Pool
            self.Pool = Pool
        else:
            self.Pool = ThreadPool
        self._json_per_worker = json_per_worker
        self._index_filter = IndexCuckooFilter(
            self.index_generator, dump_file_path=cuckoo_dump)
        self._schema_holder = SchemaReducer(
            self._index_filter, dump_file_path=schema_dump)
        self._register_graceful_exist(
            [self._index_filter, self._schema_holder])
        self._jsonl_dump = jsonl_dump
        if self._jsonl_dump is not None:
            self._jsonl_index_filter = IndexCuckooFilter(
                self.index_generator, dump_file_path='jsonl_cuckoo.pickle'
            )
            self._jsonl_saver = JsonlSaver(
                self._jsonl_index_filter,
                archieve_file_path=jsonl_dump)
            self._register_graceful_exist([self._jsonl_index_filter])

    def _register_graceful_exist(self, objs):
        def do_exit(*args):
            for p in objs:
                p.exit_gracefully(*args)
            sys.exit(1)
        signal.signal(signal.SIGINT, do_exit)
        signal.signal(signal.SIGTERM, do_exit)

    @abc.abstractmethod
    def index_generator(self) -> typing.Iterable[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_url(self, index: str) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def is_valid_json(self, json_dict: typing.Dict) -> bool:
        return True

    @property
    def count(self):
        return sum(1 for _ in self.index_generator())

    def get_schema(self, verbose=True):
        """
        A pipeline for inferencing json schema from a
        json files generated from url indices.
        """
        # Get indices (ignroe already processed ones)
        indexs = list(self._index_filter.filter(self.index_generator()))
        index_name_pipe = indexs
        try:
            with ThreadPool(processes=self._api_thread_cnt) as th_exc:
                with self.Pool(processes=self._inference_worker_cnt) as pr_exc:
                    if verbose:
                        index_name_pipe = tqdm.tqdm(
                            index_name_pipe,
                            desc='in-dkg-flow'
                        )
                    # Mapping index to URL
                    url_index_name_pipe = map(
                        lambda index: (
                            self.get_url(index),
                            index),
                        index_name_pipe)

                    # Download Json from URL
                    json_index_name_pipe = th_exc.imap_unordered(
                        APIInferenceEngine._th_run, url_index_name_pipe)

                    if verbose:
                        json_index_name_pipe = tqdm.tqdm(
                            json_index_name_pipe, total=len(indexs),
                            desc='json-flow')

                    # Remove errorneous Json
                    json_index_name_pipe = self.filter_errorneous_json(
                        json_index_name_pipe)

                    # Saving json into jsonl file
                    if self._jsonl_dump is not None:
                        json_index_name_pipe = self._jsonl_saver.save(
                            json_index_name_pipe)

                    # Convert to batch-wise pipe
                    json_index_name_batch_pipe = InferenceEngine._batchwise_generator(
                        json_index_name_pipe, batch_size=self._json_per_worker)

                    # Inferencing Json schemas from Json Batches
                    json_schema_indexs_pipe = pr_exc.imap_unordered(
                        APIInferenceEngine._pr_run, json_index_name_batch_pipe)

                    if verbose:
                        json_schema_indexs_pipe = tqdm.tqdm(
                            json_schema_indexs_pipe, total=round(
                                len(indexs) / self._json_per_worker),
                            desc='schema-batch-flow')

                    # Reducing Json Schemas into One Union Json Schema
                    self._schema_holder.reduce(json_schema_indexs_pipe)
            return self._schema_holder.union_schema
        except BaseException as e:
            raise e
        finally:
            # Saving the final schema and process record as Pickles
            self._schema_holder.save()
            self._index_filter.save()
            if self._jsonl_dump is not None:
                self._jsonl_index_filter.save()

    @staticmethod
    def _th_run(instance):
        url, index = instance
        json_result = APIInferenceEngine._get_json(url)
        return json_result, index

    @staticmethod
    def _get_json(url):
        result = requests.get(url).json()
        return result

    def filter_errorneous_json(
            self, json_index_name_pipe: typing.Iterable[typing.Tuple[typing.Dict, str]]):
        def remove_index_of_errorneous_json(instance):
            json_result, index = instance
            if not self.is_valid_json(json_result):
                self._index_filter.remove(index)
            return json_result, index
        json_index_name_pipe = map(
            remove_index_of_errorneous_json,
            json_index_name_pipe)
        json_index_name_pipe = filter(
            lambda x: self.is_valid_json(x[0]), json_index_name_pipe)
        return json_index_name_pipe

    @staticmethod
    def _pr_run(
            json_index_name_batch: typing.List[typing.Tuple[typing.Dict, str]]):
        json_batch = [x[0] for x in json_index_name_batch]
        index_name_batch = [x[1] for x in json_index_name_batch]
        json_schema = InferenceEngine.get_schema(json_batch)
        return json_schema, index_name_batch


class IndexCuckooFilter:
    """
    Filter out index whose json schema
    has already been captured in the downstream pipeline.

    Args:
        - index_gen_builder: a function that produce a index generator which help IndexCuckooFilter to
            construct the index set.
        - dump_file_path: the path to store the index set status information.
        - error_rate: the error rate of identifying an non-existing item in the set.
    """

    def __init__(self, index_gen_builder=typing.Callable[[
    ], typing.Iterable], dump_file_path='cuckoo.pickle', error_rate: float = 0.01):
        self._dump_file_path = dump_file_path
        if os.path.exists(self._dump_file_path):
            self._cuckoo = self.load()
        else:
            # build the cuckoo filter from scratch
            indexs = list(index_gen_builder())
            assert error_rate <= 1. and error_rate > 0.
            bucket_size = round(- math.log10(error_rate)) + 1
            # fingerprint_size = int(math.ceil(math.log(1.0 / error_rate, 2) + math.log(2 * bucket_size, 2)))
            if bucket_size == 1:
                alpha = 0.5
            elif bucket_size >= 2 and bucket_size < 4:
                alpha = 0.84
            elif bucket_size >= 4 and bucket_size < 8:
                alpha = 0.95
            elif bucket_size >= 8:
                alpha = 0.98
            capacity = round(len(indexs) / alpha)
            logging.info('capacity of cuckoo filter:', capacity)
            logging.info(
                'false positive error rate of cuckoo filter:',
                error_rate)
            logging.info('bucket size of cuckoo filter:', bucket_size)
            from cuckoo.filter import CuckooFilter
            self._cuckoo = CuckooFilter(
                capacity=capacity,
                error_rate=error_rate,
                bucket_size=bucket_size)
            for index in indexs:
                self._cuckoo.insert(index)

    def remove(self, index: str):
        self._cuckoo.delete(index)

    def filter(self, index_gen: typing.Iterable):
        for index in index_gen:
            if self._cuckoo.contains(index):
                yield index

    def load(self):
        with open(self._dump_file_path, 'rb') as handle:
            result = pickle.load(handle)
        print(f'{self._dump_file_path} Loaded')
        return result

    def exit_gracefully(self, *args):
        self.save()
        print('[PackageCuckooFilter] exit gracefully')

    def save(self):
        with open(self._dump_file_path, 'wb') as handle:
            pickle.dump(self._cuckoo, handle, protocol=pickle.HIGHEST_PROTOCOL)
        print(f'{self._dump_file_path} Saved')


class JsonlSaver:
    """
    This class enable saving the json(s) into a archive jsonl file.
    """

    def __init__(self, cuckoo_filter: IndexCuckooFilter,
                 archieve_file_path='archieve.jsonl'):
        self._cuckoo_filter = cuckoo_filter
        self._archieve_file_path = archieve_file_path

    def save(
            self, json_index_producer: typing.Iterable[typing.Tuple[dict, str]]):
        """
        Check and save the passing index-json tuple
        """
        with open(self._archieve_file_path, 'w') as f:
            for json, index in json_index_producer:
                if self._cuckoo_filter._cuckoo.contains(index):
                    f.write(json_package.dumps(json))
                    f.write('\n')
                    self._cuckoo_filter.remove(index)
                else:
                    pass
                yield json, index


class SchemaReducer:
    """
    This class reduces the json schemas produced by
    the inference_workers.

    It stored the union schema of the inferenced json schemas
    and captured the record the corresponding indices
    in the IndexCuckooFilter.
    """

    def __init__(self, cuckoo_filter: IndexCuckooFilter,
                 dump_file_path='schema.pickle'):
        self._cuckoo_filter = cuckoo_filter
        self._dump_file_path = dump_file_path
        if os.path.exists(self._dump_file_path):
            self._current_schema = self.load()
        else:
            self._current_schema = None

    def reduce(
            self, schema_indices_producer: typing.Iterable[typing.Tuple[JsonSchema, typing.List[str]]]):
        for schema, indices in schema_indices_producer:
            for index in indices:
                self._cuckoo_filter.remove(index)
            if self._current_schema is not None:
                self._current_schema |= schema
            else:
                self._current_schema = schema

    def load(self):
        with open(self._dump_file_path, 'rb') as handle:
            result = pickle.load(handle)
        print(f'{self._dump_file_path} Loaded')
        return result

    def save(self):
        with open(self._dump_file_path, 'wb') as handle:
            pickle.dump(
                self._current_schema,
                handle,
                protocol=pickle.HIGHEST_PROTOCOL)
        print(f'{self._dump_file_path} Saved')

    def exit_gracefully(self, *args):
        self.save()
        print('[SchemaReducer] exit gracefully')

    @property
    def union_schema(self):
        return self._current_schema



from .basic import JsonSchema, Atomic, Union, Optional, Unknown
from .records import Record, UniformRecord, DynamicRecord
from .array import Array


__all__ = [
    'JsonSchema',
    'Unknown',
    'Atomic',
    'Union',
    'Optional',
    'Array',
    'Record',
    'UniformRecord',
    'DynamicRecord'
]



"""
A basic json schema inference engine
"""
import typing
from ..fitter import fit
from ..objs import JsonSchema
from .reduce import reduce_schema


__all__ = ['InferenceEngine']


class InferenceEngine:
    def __init__(self, batch_size=100):
        self._batch_size = batch_size

    def get_schema_iteratively(self, json_pipe: typing.Iterable[typing.Any]):
        batch_pipe = InferenceEngine._batchwise_generator(
            json_pipe, batch_size=self._batch_size)
        schema_pipe = map(
            lambda batch: InferenceEngine.get_schema(batch),
            batch_pipe)
        return reduce_schema(schema_pipe)

    @staticmethod
    def get_schema(json_batch: typing.List[typing.Any]) -> JsonSchema:
        return reduce_schema(
            map(fit, json_batch))

    @staticmethod
    def _batchwise_generator(gen, batch_size=100):
        batch = []
        for i, element in enumerate(gen):
            batch.append(element)
            if i % batch_size == (batch_size - 1):
                yield batch
                del batch
                batch = []
        if batch:
            yield batch



"""
Array Json Schema Objects

e.g., [1,2,3,'apple']
"""
import copy
from .basic import JsonSchema
__all__ = [
    'Array'
]


class Array(JsonSchema):
    def __init__(self, content: JsonSchema):
        super().__init__(content)

    def check_content(self):
        assert isinstance(
            self._content, JsonSchema), 'Array content should be JsonSchema'

    def __repr__(self):
        return f'Array({self._content})'

    def __or__(self, e):
        if isinstance(e, Array):
            new = copy.deepcopy(e)
            old = copy.deepcopy(self)
            return Array(old._content | new._content)
        else:
            return self._base_or(e)



"""
Basic Json Schema Objects
"""
import abc
import copy
import typing
__all__ = [
    'Atomic',
    'Union',
    'Optional',
    'Unknown'
]


class JsonSchema:
    def __init__(self, content=None):
        self._content = content
        self.check_content()

    @abc.abstractmethod
    def check_content(self):
        raise NotImplementedError

    def __eq__(self, e):
        if isinstance(self, type(e)):
            return self._content == e._content
        else:
            return False

    def __or__(self, e):
        return self._base_or(e)

    def _base_or(self, e):
        old = copy.deepcopy(self)
        new = copy.deepcopy(e)
        if old == new or isinstance(e, Unknown):
            return old
        else:
            if old._content is None:
                if new._content is None:
                    return old
                else:
                    if isinstance(new, Optional):
                        return new
                    else:
                        return Optional(new)
            elif new._content is None:
                return Optional(old)
            elif isinstance(new, Union):
                new |= old
                return new
            else:
                return Union({old, new})

    def _base_hash(self):
        return hash(self._content)

    def __hash__(self):
        return self._base_hash()


class Atomic(JsonSchema):
    """
    simple json units, such as `int`, `float`, `null`,
        `str`, etc.
    """

    def __init__(self, content: typing.Union[type, None]):
        super().__init__(content)

    def check_content(self):
        assert isinstance(
            self._content, type) or self._content is None, 'Atomic content should be `int`, `str`, `float`, or `None`'

    def __repr__(self):
        if self._content is None:
            return 'Atomic(None)'
        else:
            return f'Atomic({self._content.__name__})'


class Unknown(JsonSchema):
    def check_content(self):
        pass

    def __or__(self, e):
        return e

    def __repr__(self):
        return 'Unknown()'


class Union(JsonSchema):

    def __init__(self, content: typing.Set[JsonSchema]):
        super().__init__(content)

    def check_content(self):
        assert isinstance(self._content, set), 'Union content should be set'
        for e in self._content:
            assert isinstance(
                e, JsonSchema), 'Union content elements should be JsonSchema'

    def __repr__(self):
        content_str = ', '.join(map(str, list(self._content)))
        content_str = '{' + content_str + '}'
        return f'Union({content_str})'

    def __or__(self, e):
        new = copy.deepcopy(e)
        if isinstance(new, Union):
            new_set = copy.deepcopy(self._content)
            for element in new._content:
                new_set |= element
            return Union(new_set)
        else:
            new_set = copy.deepcopy(self._content)
            new_set.add(new)
            return Union(new_set)

    def __hash__(self):
        return hash(tuple(self._content))


class Optional(Union):
    def __init__(self, content: JsonSchema):
        self._the_content = content
        super().__init__({Atomic(None), content})

    def __repr__(self):
        return f'Optional({self._the_content})'

    def __or__(self, e: JsonSchema):
        old = copy.deepcopy(self)
        new = copy.deepcopy(e)
        if new._content is None or old == new or old._the_content == new:
            return old
        else:
            # add element in new to the orignal element in OptionalUnion
            if isinstance(new, Union):
                if isinstance(new, Optional):
                    result = Optional(old._the_content | new._the_content)
                else:
                    result = Optional(old._the_content | new)
            else:
                result = Optional(old._the_content | new)
            return result



import abc
import os
import subprocess
from threading import Thread
import signal
from . import remote
from ..schema.inference.reduce import reduce_schema


__all__ = ['JsonlInferenceEngine']


def get_schema_remotely(jsonl_path, verbose=True, position=0, batch_size=1000):
    import json
    import tqdm
    from jsonschema_inference.schema import InferenceEngine
    total = sum(1 for _ in open(jsonl_path, 'r'))
    with open(jsonl_path, 'r') as f:
        json_pipe = map(json.loads, f)
        if verbose:
            json_pipe = tqdm.tqdm(
                json_pipe, total=total, desc=jsonl_path, position=position)
        schema = InferenceEngine(
            batch_size=batch_size).get_schema_iteratively(json_pipe)
    return schema


class JsonlInferenceEngine:
    """
    Args:
        - inference_worker_cnt: number of processes inferencing the json schema
        - tmp_dir: the path to store the splitted jsonl files.
    Methods to be overide:
        - jsonl_path: path to the jsonl file.
    """

    def __init__(self, inference_worker_cnt=8, tmp_dir='/tmp'):
        self._inference_worker_cnt = inference_worker_cnt
        self._tmp_dir = tmp_dir
        # NOTE: Whether the engine is pypy or python depends on how the code is
        # executed
        self._engine = 'pypy'  # or python
        if inference_worker_cnt > 1:
            signal.signal(signal.SIGTERM, self._graceful_exit)
            signal.signal(signal.SIGINT, self._graceful_exit)

    @abc.abstractproperty
    def jsonl_path(self):
        raise NotImplementedError

    def get_schema(self, verbose=True):
        if self._inference_worker_cnt == 1:
            result = get_schema_remotely(self.jsonl_path, verbose=verbose)
        else:
            result = self.get_schema_parallel(verbose=verbose)
        return result

    def get_schema_parallel(self, verbose=True):
        self._split_jsonl(self._inference_worker_cnt)
        schemas = []
        self._remote_gateways = []

        def layered_get_schema(i, filename):
            if self._engine == 'pypy':
                gw, decorated_get_schema = remote.pypy(get_schema_remotely)
            else:
                gw, decorated_get_schema = remote.python(get_schema_remotely)
            self._remote_gateways.append(gw)
            schema = decorated_get_schema(
                filename, verbose=verbose, position=i)
            schemas.append(schema)
        try:
            split_file_paths = os.listdir(self._split_path)
            split_file_pipe = map(
                lambda file: os.path.join(
                    self._split_path,
                    file),
                split_file_paths)
            # construct the threads
            self.threads = [Thread(target=layered_get_schema, args=(i, filename))
                            for i, filename in enumerate(split_file_pipe)]
            # start the threads
            for thread in self.threads:
                thread.daemon = False
                thread.start()
            # wait for the threads to complete
            for i, thread in enumerate(self.threads):
                thread.join()
                th = self.threads[i]
                self.threads[i] = None
                del th
            result = reduce_schema(schemas)
            return result
        except BaseException as e:
            raise e
        finally:
            self._exit()

    def _exit(self):
        self.__stop_gateways()
        self.__remove_split_files()
        print('exit')

    def _graceful_exit(self, signal=None, frame=None):
        self._exit()
        print('graceful exited')
        os._exit(0)

    def __stop_gateways(self):
        for gw in self._remote_gateways:
            try:
                gw.exit()
            except BaseException:
                pass
        print('remote gateways stopped')

    def __remove_split_files(self):
        if os.path.exists(self._split_path):
            subprocess.run(['rm', '-r', self._split_path])
            print('split files removed')

    def _split_jsonl(self, split_count):
        out = subprocess.check_output(['wc', '-l', self.jsonl_path])
        total = int(out.decode("utf-8").split(self.jsonl_path)[0])
        if not os.path.exists(self._split_path):
            os.mkdir(self._split_path)
        json_cnt_per_file = int(total / split_count)
        subprocess.run(['split', '-l', str(json_cnt_per_file),
                       self.jsonl_path, self._split_path + '/'])
        return json_cnt_per_file

    @property
    def _split_path(self):
        return os.path.join(self._tmp_dir, 'split')



import typing
from functools import reduce
from ..objs.basic import JsonSchema
from ..objs.basic import Unknown


def reduce_schema(json_schemas: typing.Iterable[JsonSchema]) -> JsonSchema:
    if json_schemas:
        result = reduce(lambda a, b: a | b, json_schemas)
    else:
        result = Unknown()
    return result



"""
How to sent this package onto PyPi?

1) Building Package Release Tar:
```python setup.py sdist```

2) Upload Package to PyPi:
```pip install twine```
```twine upload dist/*```
"""
import pathlib
import setuptools
from setuptools import find_packages

with open("README.md", "r") as fh:

    long_description = fh.read()


with open(str(pathlib.Path(__file__).parent.absolute()) +
          "/jsonschema_inference/version.py", "r") as fh:
    version = fh.read().split("=")[1].replace("'", "")

setuptools.setup(

    name="jsonschema-inference",

    version=version,

    author="jeffreylin",

    author_email="jeffrey82221@gmail.com",

    description="Fast Json Schema Inference Engine for API and Jsonl",

    long_description=long_description,

    long_description_content_type="text/markdown",

    url="https://github.com/jeffrey82221/jsonschema_inference",

    packages=find_packages(exclude=('tests',)),

    classifiers=[

        "Programming Language :: Python :: 3",

        "License :: OSI Approved :: MIT License",

        "Operating System :: OS Independent",

    ],
    python_requires='>=3.7',
    tests_require=['pytest'],
    install_requires=[
        'execnet',
        'requests',
        'scalable-cuckoo-filter',
        'tqdm',
        'urllib3',
        'autopep8'
    ],
    entry_points={
        'console_scripts': [
            'jsonschema-inference = \
        jsonschema_inference.cmd.inference:run',
        ],
    },
    extras_require={
        'ray': ["ray"]
    }
)
